#!/usr/bin/env python3
# Basic python includes
import random
import time
import numpy as np
np.set_printoptions(precision=2)
import numba as nb
from numba import njit, cuda

A = 4
t = 4
tp1 = t+1
D = 96
L = 6
assert D%L == 0
DL = D//L
threads = t*DL

@cuda.jit(fastmath=True)
def gpu_sketch(global_hashes, global_signs, seq, starts, T):

    seqid = cuda.blockIdx.x
    start = starts[seqid]
    end = starts[seqid+1]

    l = cuda.threadIdx.x
    k = cuda.threadIdx.y
    assert k < t
    assert l < D

    # Copy the global hashes/signs to shared memory.
    hashes = cuda.shared.array(shape=(t, A), dtype=nb.int32)
    signs = cuda.shared.array(shape=(t, A), dtype=nb.float32)
    if l < A:
        hashes[k][l] = global_hashes[k][l]
        hashes[k][l] = global_hashes[k][l]

    # Shared memory for temporary DP tables.
    Tin = cuda.shared.array(shape=(2, tp1, D), dtype=nb.float32)


    Tin[0][k][l] = 0
    Tin[0][k+1][l] = 0
    Tin[1][k][l] = 0
    Tin[1][k+1][l] = 0

    cuda.syncthreads()

    Tin[0][0][0] = 1
    Tin[1][0][0] = 1

    cuda.syncthreads()

    j = 0

    local_seq = cuda.shared.array(shape=threads, dtype=nb.int32)
    # Loop over characters in the sequence.
    tid = l + k * DL
    for i in range((end-start)//threads):
        idx = start + i*threads + tid
        local_seq[tid] = seq[idx]
        cuda.syncthreads()

        for c in local_seq:
            h = hashes[c][k]
            s = signs[c][k]
            # Compute the shifted target index, avoiding a modulo operation.
            for ll in range(L*l, L*(l+1)):
                r = ll + h
                r -= D if r >= D else 0
                # Write to output tensor.
                Tin[1-j][k+1][ll] = Tin[j][k+1][ll] + s * Tin[j][k][r]
            j = 1-j
            cuda.syncthreads()

    # TODO: Handle leftover part of sequence

    # Copy to result.
    #for ll in range(L*l, L*(l+1)):
    for ll in range(l, D, DL):
        T[cuda.blockIdx.x][k+1][ll] = Tin[j][k+1][ll]

class GTS:
    def __init__(self):
        random.seed(31415)
        # A t*D array of random integers in [0, D)
        self.hashes = np.array([[random.randrange(0, D) for c in range(A)] for h in range(t)], dtype=np.int32)
        # A t*D array of random +-1
        self.signs  = np.array([[random.randrange(-1, 2, 2) for c in range(A)] for h in range(t)], dtype=np.int32)

        self.d_hashes = cuda.to_device(np.array(self.hashes, dtype=np.int32))
        self.d_signs = cuda.to_device(np.array(self.signs, dtype=np.float32))

    def sketch(self, seqs):
        # Allocate memory for the response.
        stream = cuda.stream()

        blocks = len(seqs)

        seq = np.concatenate(seqs)
        starts = np.array(np.cumsum(np.array([0] + [len(seq) for seq in seqs]), dtype=np.int32), dtype=np.int32)

        d_seq = cuda.to_device(seq, stream=stream)
        d_starts = cuda.to_device(starts, stream=stream)
        d_T = cuda.device_array((blocks, t+1, D), dtype=np.float32, stream=stream)

        # One thread per (l, k) <= (D, t)
        gpu_sketch[(blocks,1), (DL, t), stream](self.d_hashes, self.d_signs, d_seq, d_starts, d_T)

        return (d_starts, d_seq, d_T)

# Generate kernels * blocks_per_kernel sequences of given length.
blocks_per_kernel = 20
kernels = 100
length = 30000
seqs = [[np.array(random.choices([0,1,2,3], k=length), dtype=np.int8) for b in range(blocks_per_kernel)] for k in range(kernels)]

# Initialize wrapper classes
g = GTS()

# Run once for precompilation
results = [g.sketch(block) for block in seqs]
cuda.synchronize()

# Timed and profiled runs.
# nvvp --profile-from-start off -fo profile.nvvp ./experiment.py
cuda.profile_start()

s = time.time()
results = [g.sketch(block) for block in seqs]
cuda.synchronize()
e = time.time()
print('Concatenation took: ', e-s)

cuda.profile_stop()
